@@ -1,9 +1,10 @@
 # -*- coding: utf-8 -*-
-# (c) 2020 Martin Wendt and contributors; see https://github.com/mar10/yabs
+# (c) 2020-2022 Martin Wendt and contributors; see https://github.com/mar10/yabs
 # Licensed under the MIT license: https://www.opensource.org/licenses/mit-license.php
 """
 """
 import logging
+import math
 import os
 import re
 import sys
@@ -15,11 +16,14 @@ from io import StringIO
 from pathlib import Path
 from shutil import rmtree
 from threading import Event, RLock, Thread
+from typing import List, Tuple, Union
 
 from snazzy import Snazzy, emoji, gray, green, red, yellow
 
 logger = logging.getLogger("yabs")
 
+CONFIG_NAME = "yabs.yaml"
+
 
 class YabsError(RuntimeError):
     """Base class for all exception that we deliberatly throw."""
@@ -37,13 +41,81 @@ class CheckError(ExitingYabsError):
 
 
 class ConfigError(ExitingYabsError):
-    """Invalid yabs.yaml or command line (terninates without stacktrace)."""
+    """Invalid yabs.yaml or command line (terminates without stacktrace)."""
 
 
 class NO_DEFAULT:
     """Used as default parameter to distinguish from `None`."""
 
 
+def get_folder_file_names(folder):
+    """Return folder files names as set."""
+    p = Path(folder)
+    name_set = set([e.name for e in p.iterdir()])
+    return name_set
+
+
+class FolderContentMonitor:
+    def __init__(
+        self, artifacts_def: Union[dict, None], *, create_folder: bool = False
+    ) -> None:
+        self.artifacts_def = artifacts_def
+        self.path = None
+        self.create_folder = create_folder
+        self.prev_files = None
+        self.cur_files = None
+        self.added_files = None
+        self.changed_or_added_files = None
+        self.changed_or_added_by_tag = None
+        if artifacts_def:
+            self.path = Path(artifacts_def["folder"]).absolute()
+
+    def __enter__(self):
+        path = self.path
+        if not path:
+            return self
+
+        if not path.is_dir():
+            if not self.create_folder:
+                raise ValueError(f"Folder not found: {path}")
+            # elif self.dry_run:
+            #     log_dry(f"Creating dist folder: {path}")
+            else:
+                log_info(f"Creating dist folder: {path}")
+                path.mkdir()
+
+        self.prev_stats = {}
+        for e in path.iterdir():
+            self.prev_stats[e.name] = e.stat()
+        self.prev_files = set(self.prev_stats.keys())
+
+        return self
+
+    def __exit__(self, exc_type, exc_value, traceback):
+        if not self.path:
+            return
+
+        self.added_files = set()
+        self.changed_or_added_files = set()
+        self.changed_or_added_by_tag = {}
+        for e in self.path.iterdir():
+            cur_stat = e.stat()
+            prev_stat = self.prev_stats.get(e.name)
+            if prev_stat is None:
+                self.added_files.add(e.name)
+                self.changed_or_added_files.add(e.name)
+            elif cur_stat.st_mtime != prev_stat.st_mtime:
+                self.changed_or_added_files.add(e.name)
+
+        for fspec in self.changed_or_added_files:
+            for tag, pattern in self.artifacts_def["matches"].items():
+                # print("MATCH", tag, pattern, fspec)
+                if re.match(pattern, fspec):
+                    full_path = (self.path / fspec).absolute()
+                    self.changed_or_added_by_tag[tag] = full_path
+        return
+
+
 def init_logging(verbose=3, path=None):
     """CLI calls this."""
     if verbose < 1:
@@ -70,11 +142,11 @@ def init_logging(verbose=3, path=None):
 
     if path:
         if os.path.isdir(path):
-            fname = "release-tool_{}.log".format(timetag())
+            fname = f"release-tool_{timetag()}.log"
             path = os.path.join(path, fname)
-        logger.info("Writing log to '{}'".format(path))
+        logger.info(f"Writing log to '{path}'")
         if os.path.isfile(path):
-            logger.warning("Removing log file '{}'".format(path))
+            logger.warning(f"Removing log file '{path}'")
             os.remove(path)
         hdlr = logging.FileHandler(path)
         formatter = logging.Formatter(
@@ -90,7 +162,7 @@ def init_logging(verbose=3, path=None):
         logger.addHandler(hdlr)
 
     # Silence requests `InsecureRequestWarning` messages
-    if verbose < 3:
+    if verbose <= 3:
         warnings.filterwarnings("ignore", message="Unverified HTTPS request")
     else:
         warnings.filterwarnings("once", message="Unverified HTTPS request")
@@ -110,29 +182,11 @@ def check_cli_verbose(default=3):
     return verbose
 
 
-# TODO: Can be removed with latest version of Snazzy:
-ANSI_ESCAPE_8BIT_STR = re.compile(
-    r"(?:\x1B[@-Z\\-_]|[\x80-\x9A\x9C-\x9F]|(?:\x1B\[|\x9B)[0-?]*[ -/]*[@-~])"
-)
-ANSI_ESCAPE_8BIT_BYTES = re.compile(
-    br"(?:\x1B[@-Z\\-_]|[\x80-\x9A\x9C-\x9F]|(?:\x1B\[|\x9B)[0-?]*[ -/]*[@-~])"
-)
-
-
-def cleanup_ansi_codes(s):
-    # 7-bit and 8-bit C1 ANSI sequences
-    if isinstance(s, str):
-        res = ANSI_ESCAPE_8BIT_STR.sub("", s)
-    else:
-        res = ANSI_ESCAPE_8BIT_BYTES.sub(b"", s)
-    return res
-
-
 _prefix_map = None
 _prefix_map_valid = False
 
 
-def write(msg, level="info", prefix=False, output=None, output_level=None):
+def write(msg: str, *, level="info", prefix=False, output=None, output_level=None):
     """"""
     # We want to cache the color formatted strings, however the
     # `snazzy` formatter may not be initialized yet.
@@ -169,7 +223,7 @@ def write(msg, level="info", prefix=False, output=None, output_level=None):
     logger.log(level, prefix + msg)
 
     if output:
-        prefix_len = len(cleanup_ansi_codes(prefix))
+        prefix_len = len(Snazzy.cleanup(prefix))
         prefix = (" " * prefix_len) + " > "
         lines = output.split("\n")
         # strip trailing empty lines
@@ -185,31 +239,31 @@ def write(msg, level="info", prefix=False, output=None, output_level=None):
 
 
 def log_debug(msg):
-    return write(msg, "debug", False)
+    return write(msg, level="debug", prefix=False)
 
 
 def log_info(msg):
-    return write(msg, "info", False)
+    return write(msg, level="info", prefix=False)
 
 
 def log_dry(msg):
-    return write("DRY-RUN " + msg, "info", False)
+    return write("DRY-RUN " + msg, level="info", prefix=False)
 
 
 def log_response(info, output, level, dry_run):
-    return write(info, level, True, output=output)
+    return write(info, level=level, prefix=True, output=output)
 
 
 def log_ok(msg):
-    return write(msg, "info", True)
+    return write(msg, level="info", prefix=True)
 
 
 def log_warning(msg):
-    return write(msg, "warning", True)
+    return write(msg, level="warning", prefix=True)
 
 
 def log_error(msg):
-    return write(msg, "error", True)
+    return write(msg, level="error", prefix=True)
 
 
 def set_console_ctrl_handler(
@@ -236,7 +290,7 @@ def set_console_ctrl_handler(
         kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)
         logger.info("Loaded kernel32 DLL on Windows.")
     except Exception as e:
-        logger.warning("Could not load kernel32 DLL on Windows: {}".format(e))
+        logger.warning(f"Could not load kernel32 DLL on Windows: {e}")
         return False
 
     CTRL_C_EVENT = 0
@@ -312,7 +366,7 @@ def _check_arg(argument, types, condition, accept_none):
         )
 
 
-def check_arg(argument, allowed_types, condition=NO_DEFAULT, or_none=False):
+def check_arg(argument, allowed_types, condition=NO_DEFAULT, *, or_none=False):
     """Check if `argument` has the expected type and value.
 
     **Note:** the exception's traceback is manipulated, so that the back frame
@@ -357,13 +411,24 @@ def check_arg(argument, allowed_types, condition=NO_DEFAULT, or_none=False):
         raise e.with_traceback(back_tb)
 
 
-def to_list(obj):
+def to_list(obj, or_none=False):
     """Convert a single object to a list."""
     if obj is None:
-        return []
+        return None if or_none else []
     elif isinstance(obj, (list, tuple)):
         return obj
-    return [obj]
+    elif isinstance(obj, set):
+        return list(obj)
+    return [obj]  # may ba a str
+
+
+def to_set(obj, or_none=False):
+    """Convert a single object to a set."""
+    if obj is None and or_none:
+        return None
+    if isinstance(obj, set):
+        return obj
+    return set(to_list(obj))
 
 
 def get_dict_attr(d, key_path, default=NO_DEFAULT):
@@ -404,22 +469,46 @@ def get_dict_attr(d, key_path, default=NO_DEFAULT):
             seg = seg[1:-1]
             value = value[int(seg)]
         else:
-            # raise ValueError("Segment '{}' cannot be nested".format(seg))
+            # raise ValueError(f"Segment '{seg}' cannot be nested")
             try:
                 value = getattr(value, seg)
             except AttributeError:
-                raise  # ValueError("Segment '{}' cannot be nested".format(seg))
+                raise  # ValueError(f"Segment '{seg}' cannot be nested")
 
     return value
 
 
-def timetag(seconds=True, ms=False):
+def check_dict_keys(
+    d: dict, *, known: set, mandatory: set, prefix: str, key_prefix: str = ""
+) -> List[str]:
+    """Validate a dict fo missing or unknown keys."""
+    used = set(d.keys())
+    known = known.union(mandatory)
+    missing = mandatory.difference(used)
+
+    errors = []
+
+    if missing:
+        s1 = ", ".join(sorted(f"`{key_prefix}{s}`" for s in missing))
+        errors.append(f"{prefix}Missing mandatory option(s): {s1}")
+
+    invalid = used.difference(known)
+    if invalid:
+        unused = known.difference(used)
+        s1 = ", ".join(sorted(f"`{key_prefix}{s}`" for s in invalid))
+        s2 = "'" + "', '".join(sorted(unused)) + "'"
+        errors.append(f"{prefix}Unsupported option(s): {s1} (did you mean {s2} ?)")
+
+    return errors
+
+
+def timetag(seconds=True, *, ms=False):
     """Return a time stamp string that can be used as (part of a) filename (also sorts well)."""
     now = datetime.now()
     if ms or seconds:
         s = now.strftime("%Y%m%d_%H%M%S")
         if ms:
-            s = "{}_{}".format(s, now.microsecond)
+            s = f"{s}_{now.microsecond}"
     else:
         s = now.strftime("%Y%m%d_%H%M")
     return s
@@ -433,33 +522,59 @@ def resolve_path(root, path, must_exist=True, check_root=False):
         path = os.path.join(root, path)
     path = os.path.abspath(path)
     if check_root and not path.startswith(root):
-        raise ValueError("Path must be in or below {}: {}".format(root, path))
+        raise ValueError(f"Path must be in or below {root}: {path}")
     if must_exist and not os.path.isfile(path):
-        raise ValueError("File not found: {}".format(path))
+        raise ValueError(f"File not found: {path}")
     return path
 
 
+def search_file_upward(
+    root: Union[str, Path], filename: str, *, max_level=0, or_none=False
+) -> Path:
+    """."""
+    root = Path(root).expanduser().absolute()
+    if root.is_file():
+        return root
+    assert root.is_dir(), root
+
+    target = f"{root}/{filename}"
+    cur_level = 0
+    parents = [root] + list(root.parents)
+    for parent in parents:
+        fspec = parent / filename
+        # print(f"Searching for {fspec}...")
+        if fspec.is_file():
+            return fspec
+        cur_level += 1
+        if max_level and cur_level >= max_level:
+            break
+
+    if not or_none:
+        raise FileNotFoundError(f"{target}")
+    return None
+
+
 def remove_directory(path, content_only=False, log=None):
     check_arg(path, (str, Path))
     check_arg(content_only, bool)
 
     path = Path(path).absolute().resolve()
     if not path.is_dir():
-        raise ValueError("Not a directory: {}".format(path))
+        raise ValueError(f"Not a directory: {path}")
 
     if content_only:
         for p in path.iterdir():
             if p.is_file():
                 if log:
-                    log("Removing file {}".format(p))
+                    log(f"Removing file {p}")
                 p.unlink()
             elif p.is_dir():
                 if log:
-                    log("Removing directory {}".format(p))
+                    log(f"Removing directory {p}")
                 rmtree(p)
     else:
         if log:
-            log("Removing directory {}".format(path))
+            log(f"Removing directory {path}")
         rmtree(path)
     return
 
@@ -504,12 +619,31 @@ def shorten_string(long_string, max_chars, min_tail_chars=0, place_holder="[...]
     return long_string
 
 
-def datetime_to_iso(dt=None, microseconds=False):
+def plural_s(value) -> str:
+    """Return 's' if `value` is not singular.
+
+    Example:
+        from util import plural_s as ps
+
+        v = [1, 2]
+        print(f"Found value{ps(v)}: {v}")
+    """
+    if type(value) in (float, int):
+        n = value
+    elif isinstance(value, (dict, list, set, tuple)):
+        n = len(value)
+    else:  # type(value) is str:
+        # May be None, str (assume 'one word', not: 'many characters') or other
+        n = 1
+    return "" if n == 1 else "s"
+
+
+def datetime_to_iso(dt=None, *, microseconds: bool = False) -> str:
     """Return current UTC datetime as ISO formatted string."""
     if dt is None:
         dt = datetime.now()
     if not microseconds:
-        dt.replace(microsecond=0)
+        dt = dt.replace(microsecond=0)
     return dt.isoformat(sep=" ")
 
 
@@ -525,38 +659,46 @@ def datetime_to_iso(dt=None, microseconds=False):
 #     return iso_to_datetime(iso).timestamp()
 
 
-def format_elap(seconds, count=None, unit="items", high_prec=False):
+def format_elap(
+    seconds, *, count=None, unit="items", high_prec=False, short_suffix=False
+):
     """Return elapsed time as H:M:S.h string with reasonable precision."""
     days, seconds = divmod(seconds, 86400) if seconds else (0, 0)
 
     if seconds >= 3600:
         m, s = divmod(seconds, 60)
         h, m = divmod(m, 60)
+        suff = "h" if short_suffix else " hrs"
         if high_prec:
-            res = "{:d}:{:02d}:{:04.1f} hrs".format(int(h), int(m), s)
+            res = f"{int(h):d}:{int(m):02d}:{s:04.1f}{suff}"
         else:
-            res = "{:d}:{:02d}:{:02d} hrs".format(int(h), int(m), int(s))
+            res = f"{int(h):d}:{int(m):02d}:{int(s):02d}{suff}"
     elif seconds >= 60:
         m, s = divmod(seconds, 60)
+        suff = "m" if short_suffix else " min"
         if high_prec:
-            res = "{:d}:{:05.2f} min".format(int(m), s)
+            res = f"{int(m):d}:{s:05.2f}{suff}"
         else:
-            res = "{:d}:{:02d} min".format(int(m), int(s))
+            res = f"{int(m):d}:{int(s):02d}{suff}"
     else:
+        suff = "s" if short_suffix else " sec"
         if high_prec:
-            res = "{:.3f} sec".format(seconds)
+            res = f"{seconds:.3f}{suff}"
         elif seconds > 5:
-            res = "{:.1f} sec".format(seconds)
+            res = f"{seconds:.1f}{suff}"
         else:
-            res = "{:.2f} sec".format(seconds)
+            res = f"{seconds:.2f}{suff}"
 
     if days == 1:
-        res = "{} day {}".format(int(days), res)
+        suff = "d" if short_suffix else " day"
+        res = f"{int(days)}{suff} {res}"
     elif days:
-        res = "{} days {}".format(int(days), res)
+        suff = "d" if short_suffix else " days"
+        res = f"{int(days)}{suff} {res}"
 
     if count and (seconds > 0):
-        res += ", {:,.1f} {}/sec".format(float(count) / seconds, unit)
+        suff = "s" if short_suffix else "sec"
+        res += ", {:,.1f} {}/{}".format(float(count) / seconds, unit, suff)
     return res
 
 
@@ -727,6 +869,78 @@ def lstrip_string(s, prefix, ignore_case=False):
 #     return s
 
 
+def progress_bar_str(
+    progress: float,
+    *,
+    width: int = 10,
+    level: str = "info",
+    border: Union[str, Tuple[str]] = "|",  # ("[", "]"),
+    add_percentage: bool = True,
+    fill_char: str = " ",
+) -> str:
+    color_map = {
+        "ok": green,
+        "warning": yellow,
+        "error": red,
+    }
+    progress = min(1.0, max(0.0, progress))
+    real_progress = progress
+
+    if type(border) is str:
+        border = (border, border)
+    if border:
+        width = width - (len(border[0]) - len(border[1]))
+    width = max(1, int(width))
+
+    whole_width = math.floor(progress * width)
+    remainder_width = (progress * width) % 1
+    part_width = math.floor(remainder_width * 8)
+
+    part_char = [" ", "▏", "▎", "▍", "▌", "▋", "▊", "▉"][part_width]
+    if (width - whole_width - 1) < 0:
+        part_char = ""
+
+    # Very small values should always display a very small bar
+    if whole_width < 1 and part_char == " " and progress > 0:
+        part_char = "▏"
+
+    line = "█" * whole_width + part_char + fill_char * (width - whole_width - 1)
+
+    if level in color_map:
+        line = color_map[level](line)
+    if border:
+        line = f"{border[0]}{line}{border[1]}"
+    if add_percentage:
+        p100 = f"{100.0*real_progress:.1f}"
+        line += " {: >5}%".format(p100)
+    return line
+
+
+# print(progress_bar_str(0.123))
+
+# print(progress_bar_str(0.123, width=3, border=False))
+# print(progress_bar_str(0.023, width=3, border=False))
+
+# print(progress_bar_str(0, width=1, border=False))
+# print(progress_bar_str(0.023, width=1, border=False))
+# print(progress_bar_str(0.123, width=1, border=False))
+# print(progress_bar_str(0.13, width=1, border=False))
+# print(progress_bar_str(1.123, width=1, border=False))
+
+# print(progress_bar_str(0, width=3))
+# print(progress_bar_str(0.023, width=3))
+# print(progress_bar_str(0.123, width=3))
+# print(progress_bar_str(0.13, width=3))
+# print(progress_bar_str(1.123, width=3))
+
+# print(progress_bar_str(0))
+# print(progress_bar_str(0.023))
+# print(progress_bar_str(0.123))
+# print(progress_bar_str(0.13))
+# print(progress_bar_str(1.123))
+# raise
+
+
 def run_process_streamed(
     process,
     name=None,
@@ -766,9 +980,9 @@ def run_process_streamed(
     out = StringIO()
     pid = process.pid
     if name:
-        name = "<{}> {}".format(pid, name)
+        name = f"<{pid}> {name}"
     else:
-        name = "<{}>".format(pid)
+        name = f"<{pid}>"
 
     if on_output is None:
         on_output = logger.info
@@ -808,7 +1022,7 @@ def run_process_streamed(
                 return
         local_vars["last_flush"] = now
         if prefix_chunks:
-            on_output("process({}) stdout: {}".format(name, line_str))
+            on_output(f"process({name}) stdout: {line_str}")
         else:
             on_output(line_str)
         return
@@ -821,7 +1035,7 @@ def run_process_streamed(
         return
 
     # TODO: only when logger.LEVEL == DEBUG
-    flusher = Thread(target=flush_handler, name="Flush process output {}".format(name))
+    flusher = Thread(target=flush_handler, name=f"Flush process output {name}")
     flusher.start()
 
     try:
